﻿local Info = Info or package.loaded.regscript or function(...) return ... end --luacheck: ignore 113/Info
local nfo = Info { _filename or ...,
  name        = "BufferScroll";
  description = "Macros to navigate in large console buffer";
  version     = "0.7"; --http://semver.org/lang/ru/
  author      = "jd";
  url         = "http://forum.farmanager.com/viewtopic.php?f=60&t=8675";
  id          = "544A79FC-7029-4F79-977F-AA36DB57F389";
  minfarversion = {3,0,0,5067,0};
  --disabled    = false;
  options     = {
    prompt="^%a:%\\.->", --regexp for FAR prompt search
    height=10000,
    real_top=true,
    --key_copy_line="CtrlIns",
    --key_copy_screen="CtrlShiftIns",
    --key_save_buffer="CtrlAltIns",
  };
}
if not nfo or nfo.disabled then return end

local O = nfo.options
assert(O.prompt and type(O.prompt)=="string","Option 'prompt' must be string/boolean")
assert(not O.height or type(O.height)=="number","Option 'height' must be number")

local ptn_path = "^.+[\\/]"
local _path = (...):match(ptn_path)
function nfo.help()
  far.ShowHelp(_path,"Contents","FHELP_CUSTOMPATH")
end;
local function HelpTopic(s)
  return ("<%s>%s"):format(_path,s)
end

if not Far.GetConfig"System.WindowMode" then
  panel.GetUserScreen()
  io.write(nfo.name,": System.WindowMode is false, exiting")
  panel.SetUserScreen()
  return
end

-------------------------------------------------------------------------------
local allPanels = "Shell QView Tree Info Search "
local allPanels_ac = allPanels.."ShellAutoCompletion "

local function far_in_window() --vertical only
  local f = far.AdvControl"ACTL_GETFARRECT"
  local w = win.GetConsoleScreenBufferInfo()
  local yDelta = f.Top-w.WindowTop
  return yDelta==0, yDelta
end

local function inUserScreen_emptycmd()
  return (allPanels:find(Area.Current) and not APanel.Visible and CmdLine.Empty)--reserve some keys for cmdline
      or Area.Desktop
      or not far_in_window()
end
local function inUserScreen()
  return allPanels:find(Area.Current) and not APanel.Visible
      or Area.Desktop
      or not far_in_window()
end

local find_real_beginning -- fwd decl
local getOffset = {
  Up   =function()     return -1 end,
  PgUp =function(info) return -(info.WindowBottom-info.WindowTop+1) end,
  Home =function(info)
    return not O.real_top and info.WindowTop~=0 and -info.WindowTop
                                                 or find_real_beginning()
  end,
  Down =function()     return  1 end,
  PgDn =function(info) return  (info.WindowBottom-info.WindowTop+1) end,
  End  =function(info) return  info.SizeY-info.WindowBottom end,

  Left =function()     return -1 end,
  Right=function()     return  1 end,
}
getOffset.MsWheelUp   = getOffset.Up
getOffset.MsWheelDown = getOffset.Down

local NumPad={Num9="PgUp",Num3="PgDn",Num8="Up",Num2="Down",Num7="Home",Num1="End"}
local function getDelta(key,info)
  key = key:match"R?CtrlR?Alt(.+)" or key
  key = NumPad[key] or key
  info = info or win.GetConsoleScreenBufferInfo()
  if not getOffset[key] then error(("getDelta: unexpected key '%s'"):format(key),2) end
  return getOffset[key](info)
end

-------------------------------------------------------------------------------
local function scrollToFar()
  local in_window,yDelta = far_in_window()
  if not in_window then
    Far.Window_Scroll(yDelta)
    return yDelta
  end
end

Macro { description="Buffer: scroll down to Far window";
  area="Common"; key="Esc";
  priority=60;--!!
  id="BDAC4B4B-4BC3-4CAE-BAE7-118B62DB4875";
  condition=scrollToFar;
  action=function() end;--http://bugs.farmanager.com/view.php?id=2391
}

Macro { description="Buffer: scroll to Far before Ctrl+O";
  area="Common"; key="CtrlO";
  priority=60;--!!
  id="EBDFB192-943A-4DFC-919D-480BC7F29BD5";
  condition=function() scrollToFar() end;
  action=function() end;
}

-------------------------------------------------------------------------------
Macro { description="Buffer: auto-hide panels when scrolling up with CtrlAlt<Key>";
  area=allPanels_ac; key="/^LCtrlLAlt(Num[789]|(MsWheel)?Up|PgUp|Home)$/";
  priority=60;--!!
  id="DD121E14-9132-4F6E-BD00-9538111CDB5D";
  condition=function() return APanel.Visible and far_in_window() end;
  action=function() Keys"CtrlO" end;
}

Macro { description="Buffer: auto-unhide panels when scrolling down with CtrlAlt<Key>";
  area=allPanels_ac; key="/^LCtrlLAlt(Num[123]|(MsWheel)?Down|PgDn|End)$/";
  priority=60;--!!
  id="F412D64B-2D71-4FC2-82F9-E1E2822CC681";
  condition=function() return not APanel.Visible and far_in_window() end;
  action=function() Keys"CtrlO" end;
}

-------------------------------------------------------------------------------
local function akey() return mf.akey(1,1) end

if O.real_top then Macro { description="Buffer: scroll to real top";
  area=allPanels_ac; key="LCtrlAltHome LCtrlAltNum7";
  id="E8D9352E-D7C1-4168-AE3F-79499F85D013";
  action=function()
    Far.Window_Scroll(getDelta"Home")
  end;
}end

-- 0003543: Скроллинг консоли в Desktop https://bugs.farmanager.com/view.php?id=3543
Macro { description="Buffer: scroll in Desktop";
  area="Desktop"; key="/^LCtrlLAlt(Num[123789]|(MsWheel)?(Up|Down)|PgUp|PgDn|Home|End)$/";
  id="5A9E612C-D8D0-486C-8492-6439265B839C";
  action=function()
    Far.Window_Scroll(getDelta(akey()))
  end;
}

Macro { description="Buffer: scroll with mouse wheel";
  area="Common"; key="/^LCtrlLAltMsWheel(Up|Down)$/";
  priority=55;--!!
  id="7BEDA4D5-02FF-43B4-80DD-AACBEB8BF8C4";
  condition=inUserScreen;
  action=function()
    Far.Window_Scroll(getDelta(akey()))
  end;
}

Macro { description="Buffer: scroll with mouse wheel (horizontally)";
  area="Common"; key="/^RCtrlRAltMsWheel(Up|Down)|MsWheelLeft|MsWheelRight$/";
  priority=55;--!!
  id="009FDC96-E710-4483-B9ED-B3EDD82B494C";
  condition=inUserScreen;
  action=function()
    Far.Window_Scroll(getDelta(akey()),1)
  end;
}

Macro { description="Buffer: scroll with plain keys when Panels are hidden (horizontally)";
  area="Common"; key="Left Right";--??more keys (CtrlLeft/Right/...)
  priority=55;--??
  id="969C6694-0323-46DD-8F69-66AFCAAF37A9";
  condition=inUserScreen_emptycmd;
  action=function()
    Far.Window_Scroll(getDelta(akey()),1)
  end;
}

Macro { description="Buffer: scroll with plain keys when Panels are hidden";
  area="Common"; key="/^Num[123789]|(MsWheel)?(Up|Down)|PgUp|PgDn|Home|End$/";
  priority=55;--!!
  id="9D775B0E-7004-4B0F-B5DE-04C85F6EAF65";
  condition=inUserScreen_emptycmd;
  action=function()
    Far.Window_Scroll(getDelta(akey()))
  end;
}

-------------------------------------------------------------------------------
local ffi = require"ffi"
local C = ffi.C

ffi.cdef[[
//https://docs.microsoft.com/en-us/windows/console/getstdhandle
HANDLE GetStdHandle(           //WINAPI
  DWORD nStdHandle             //_In_
);
//https://docs.microsoft.com/en-us/windows/console/setconsolescreenbuffersize
BOOL SetConsoleScreenBufferSize(//WINAPI
  HANDLE hConsoleOutput,      //_In_
  COORD  dwSize               //_In_
);
//https://docs.microsoft.com/en-us/windows/console/readconsoleoutputcharacter
BOOL ReadConsoleOutputCharacterW(//WINAPI
  HANDLE hConsoleOutput,       //_In_
  LPTSTR lpCharacter,          //_Out_
  DWORD nLength,               //_In_
  COORD dwReadCoord,           //_In_
  LPDWORD lpNumberOfCharsRead  //_Out_
);
]]

local STD_OUTPUT_HANDLE = ffi.new("DWORD",-11)
local INVALID_HANDLE_VALUE = ffi.cast("HANDLE",-1)

local hConsoleOutput = C.GetStdHandle(STD_OUTPUT_HANDLE)
if hConsoleOutput==INVALID_HANDLE_VALUE then return end

if win.GetConsoleScreenBufferInfo().WindowTop==0 and O.height then
  local dwSize = ffi.new("COORD",win.GetConsoleScreenBufferInfo().SizeX,O.height)
  panel.GetUserScreen()
  C.SetConsoleScreenBufferSize(hConsoleOutput,dwSize)
  panel.SetUserScreen()
end

local wsize = ffi.sizeof"wchar_t"
local dwReadCoord = ffi.new"COORD"
local lpNumberOfCharsRead = ffi.new"DWORD[1]"
local function ReadConsoleOutputCharacter(x,y,nLength,lpCharacter)
  dwReadCoord.X,dwReadCoord.Y = x,y
  lpCharacter = lpCharacter or ffi.new("wchar_t[?]",nLength)
  local res = C.ReadConsoleOutputCharacterW(hConsoleOutput,lpCharacter,nLength,dwReadCoord,lpNumberOfCharsRead)
  if res>0 then
    return win.Utf16ToUtf8(ffi.string(lpCharacter,lpNumberOfCharsRead[0]*wsize))
  end
  far.Message(debug.getinfo(2).currentline,nil,nil,"ewl")
end

local trim_ptn = "^(.-)%s*$"
local function buff_match(start,dir,pattern,plain)
  local info = win.GetConsoleScreenBufferInfo()
  local lpCharacter = ffi.new("wchar_t[?]",info.SizeX)
  local line = start or info.WindowTop
  if dir==0 then dir = 1; line = line-1 end
  return function()
    repeat
      line = line+dir
      if line<0 or line>info.SizeY then break end --check bounds
      local str = ReadConsoleOutputCharacter(0,line,info.SizeX,lpCharacter)
      if str:find(pattern,1,plain) then
        return line,line-info.WindowTop,str:match(trim_ptn)
      end
    until false
  end
end
function find_real_beginning()
  for _,offset in buff_match(0,0,"%S") do --luacheck: ignore 512
    return offset
  end
end

-------------------------------------------------------------------------------
local last_search
Macro { description="Buffer: scroll to next match [Up or Down]";
  id="87E88108-1A74-427E-B45D-55BF68C286EF";
  area="Common";
  key="/LCtrl(Shift)?(Up|Down|Num[82])/";
  condition=inUserScreen;
  action=function()
    local key = akey()
    local dir = key:match"[p8]$" and -1 or 1
    local str = key:match"Shift" and last_search or O.prompt
    for _,offset in buff_match(nil,dir,str) do --luacheck: ignore 512
      return Far.Window_Scroll(offset)
    end
    mf.beep()
  end;
}

local Props = {
  Id=win.Uuid"2AA8DF11-FDC1-47C0-BA84-2E88EE65BA92",
  Bottom="Enter: scroll to position",
  HelpTopic=HelpTopic"",
}
local function searchBuffer(substr,plain)
  local found,min_offset,i = {},math.huge,0
  local cur
  for line,offset,str in buff_match(0,0,substr,plain) do
    i = i+1
    if offset<math.abs(min_offset) then
      min_offset = offset
      cur = i
    end
    found[i] = { text=("%4i: %s"):format(line,str), line=line, str=str}
  end
  if i==0 then mf.beep(); return end
  Props.SelectIndex = cur or i
  Props.Title = ("Search: '%s'"):format(substr)
  local item = far.Menu(Props,found)
  if item then
    far.Text() --https://bugs.farmanager.com/view.php?id=2926
    Far.Window_Scroll(item.line-far.AdvControl"ACTL_GETFARRECT".Top)
  end
  --return true
end

if O.prompt then Macro { description="Buffer: show all prompt occurencies";
  area="Common"; key="CtrlF7";
  id="BCB5ED18-81C6-44B9-9A3F-AEB9EAA185E9";
  condition=inUserScreen;
  action=function()
    searchBuffer(O.prompt)
  end;
}end

local Id = win.Uuid"1ECCA086-7C71-4853-BB84-C73E454E78A2"
local history_name = "BufferScrollInputSearchStr"
Macro { description="Buffer: input substr to search and show all matches";
  area="Common"; key=not O.prompt or "CtrlF7" or "CtrlShiftF7";
  id="3EDB5F42-85A9-4C3D-8DAE-D65E6BBE0FBE";
  condition=inUserScreen;
  action=function()
    local str = far.InputBox(Id,"Search console buffer","Enter string:",history_name,nil,nil,HelpTopic"","FIB_NONE")
    if str then
      last_search = str
      searchBuffer(str,"plain")
    end
  end;
}

-------------------------------------------------------------------------------
if O.key_copy_line then Macro { description="Buffer: copy upper line";
  area="Common"; key=O.key_copy_line; --CtrlIns
  id="74860EFE-28A5-42B7-9725-083264BBEAC9";
  condition=inUserScreen_emptycmd;
  action=function()
    local info = win.GetConsoleScreenBufferInfo()
    far.CopyToClipboard(ReadConsoleOutputCharacter(0,info.WindowTop,info.SizeX):match(trim_ptn))
  end;
}end

if O.key_copy_screen then Macro { description="Buffer: copy screen";
  area="Common"; key=O.key_copy_screen; --CtrlShiftIns
  id="FEEB0624-DBDE-41F7-882F-339082D274D2";
  condition=inUserScreen_emptycmd;
  action=function()
    local txt,i,h = {},0,Far.Height
    for _,_,str in buff_match(nil,0,"") do
      i = i+1
      txt[i] = str
      if i==h then break end
    end
    far.CopyToClipboard(table.concat(txt,"\n"))
  end;
}end

if O.key_save_buffer then Macro { description="Buffer: save to file";
  area="Common"; key=O.key_save_buffer; --CtrlAltIns
  id="E7A9F955-6A05-4712-AF25-6B8C5C8A8441";
  condition=inUserScreen_emptycmd;
  action=function()
    local outname = far.MkTemp("buffer")
    local out = assert(io.open(outname,"w"))
    for line in buff_match(0,0,"%S") do --luacheck: ignore 512
      for _,_,str in buff_match(line,0,"") do
        out:write(str,"\r\n")
      end
      break
    end
    out:close()
    editor.Editor(outname,nil,nil,nil,nil,nil,{EF_NONMODAL=1,EF_IMMEDIATERETURN=1})
  end;
}end
